const { app, BrowserWindow, ipcMain } = require("electron");
const path = require('path');
const fs = require("fs");
const crypto = require("crypto");
const http2 = require("http2");

let win;

const USE_PROD = true;
let baseURI;
let serverHost;
let connectionOptions = {};

if (USE_PROD)
{
	baseURI = `https://www.fenoxo.com/play/TiTS/release/`;
	serverHost = `https://www.fenoxo.com`;
}
else
{
	baseURI = `https://localhost:4444/`;
	serverHost = `https://localhost:4444`;
	connectionOptions.rejectUnauthorized = false;
}

/**
 * If you're reading this, yes I am aware how easy it would be to modify this to make it possible for public builds to get the bigger images.
 * If you figure it out, please keep it to yourself - the restriction isn't about punishing or keeping things back from public builds,
 * but trying to ease in to the bandwidth demand that this is going to place on our server. 
 * The larger images are a huge amount of data per user - I (Gedan) will have to spend time to engineer better solutions, time that I could
 * spend on game content...
 */
const packSizes = ["x2"];

function createWindow()
{
    win = new BrowserWindow({ 
        width: 1920,
        minWidth: 640, 
        height: 1080, 
        minHeight: 360,
		backgroundColor: "#2C3649",
		useContentSize: true,
        frame: true, 
        closable: true,
        autoHideMenuBar: true,
        webPreferences: {
            devTools: true,
            preload: path.join(__dirname, '/preload.js'),
			nodeIntegration: true
        },
		show: false
    });
    
    /*win.setIcon(path.join(__dirname, '/resources/icons/favicon.ico'));*/
	win.loadFile("index.html");

    win.on("close", () => 
	{
        win = null;
        app.exit(0);
    });

	win.once("ready-to-show", () => 
	{
		win.show();
	});
}

app.whenReady().then(() => 
{
	ipcMain.handle("acquireImages:start", handleAcquireImages);
	ipcMain.handle("acquireImages:cancel", handleCancelImages);
	ipcMain.handle("acquireImages:delete", handleDeleteImages);
	ipcMain.handle("acquireImages:checkManifest", checkImageManifest);
	ipcMain.handle("acquireImages:install", handleInstallImages);
	ipcMain.handle("acquireImages:isInstalled", isInstalled);
	ipcMain.handle("display:forceReload", forceReload);
	createWindow();
});

app.on("window-all-closed", () => 
{
    app.exit(0);
});

app.on("browser-window-focus", () => 
{
	win.webContents.send("display:triggerRedraw");
})

function forceReload()
{
	console.log("Attempting to force reload window...");
	app.relaunch();
	app.exit();
}

let downloadInProgress = false;
let cancelAcquisition = false;
let imagesRequiringUpdate = null;
let imagesRequiringRetry = null;

let session = null;
let numTransferred = 0;

/**
 * This place is a message and part of a system of messages. Pay attention to it!
 * Sending this message was important to us. We considered ourselves to be a powerful culture.
 * This place is not a place of honor, no highly esteemed deed is commemorated here. Nothing valued is here.
 * What is here was dangerous and repulsive to us. This message is a warning about danger.
 * The danger is in a particular location. It increases towards a center. The center of danger is here, of a particular size and shape, below us.
 * The danger is still present, in your time as it was in ours.
 * The danger is to the mind, and it can kill.
 * The form of the danger is an emanation of energy.
 * The danger is unleashed only if you substatially disturb this place physically. This place is best shunned and left uninhabited.
 */

function doRequest(s, target, uri)
{
	const fw = fs.createWriteStream(target);

	return new Promise((resolve, reject) => 
	{
		const path = (new URL(uri)).pathname;
		const req = s.request({ ":path": path });

		req.pipe(fw);

		fw.on("finish", () => 
		{
			fw.close();
			req.end();
			numTransferred++;
			win.webContents.send("acquireImages:notifyProgress", numTransferred);
			resolve(true);
		});

		req.on("error", (error) => 
		{
			console.log(error);

			if (!imagesRequiringRetry)
			{
				imagesRequiringRetry = [];
			}

			imagesRequiringRetry.push({ target: target, source: uri });

			reject(error);
		});
	});
}

let RECREATE_SESSION = false;

function createSession()
{
	let s = http2.connect(serverHost, connectionOptions);
	RECREATE_SESSION = true;

	s.on("error", (error) =>
	{
		console.log(error);
	});

	// This realistically needs a good way to stall out the outer loop because we can shotgun every request into a failed half-open session otherwise...
	s.on("goaway", (__code, __sID, __data) =>
	{
		if (RECREATE_SESSION)
		{
			s = http2.connect(serverHost, connectionOptions);
		}
	});

	return s;
}

function closeSession(s)
{
	RECREATE_SESSION = false;
	s.close();
}

async function handleAcquireImages()
{
	if (imagesRequiringUpdate === null || imagesRequiringUpdate.length === 0)
	{
		return false;
	}

	const sleepWait = ms => new Promise(r => setTimeout(r, ms));

	downloadInProgress = true;
	const basePath = path.join(app.getPath("userData"), "ImagePack");

	try
	{
		fs.mkdirSync(basePath);
	}
	catch (error)
	{
		if (error.code !== "EEXIST")
		{
			throw error;
		}
	}

	let requestTasks = [];
	const concurrencyLimit = 10;
	let fileId = 0;

	const chunkSize = 500;
	let chunkedList = [];

	for (let i = 0; i < imagesRequiringUpdate.length; i += chunkSize)
	{
		chunkedList.push(imagesRequiringUpdate.slice(i, i + chunkSize));
	}

	for (const chunk of chunkedList)
	{
		session = createSession();

		for (const i of chunk)
		{
			if (requestTasks.length >= concurrencyLimit)
			{
				await Promise.race(requestTasks);
			}

			console.log(`Downloading asset [${fileId++}] ${i.source}`);

			// Ensure the target folder exists

			if (cancelAcquisition)
			{
				break;
			}
			
			await fs.promises.mkdir(path.dirname(i.target), { recursive: true });
			const oneRequest = doRequest(session, i.target, i.source).then(() => 
			{
				requestTasks.splice(requestTasks.indexOf(oneRequest), 1);
			}).catch(() => 
			{
				requestTasks.splice(requestTasks.indexOf(oneRequest), 1);
			})
			requestTasks.push(oneRequest);
		}

		closeSession(session);
		while (!session.closed)
		{
			console.log("Waiting to recycle session...");
			await sleepWait(1000);
		}
	}

	console.log("Waiting for in-flight requests...");
	await Promise.all(requestTasks);	

	// If we have ANY failures, recreate the session and commence trying again
	if (imagesRequiringRetry && imagesRequiringRetry.length > 0)
	{
		closeSession(session);
		session = createSession();

		const retryCount = 2;
		for (let i = 0; i < retryCount; i++)
		{
			let failingListCopy = imagesRequiringRetry;
			imagesRequiringRetry = [];

			if (!Array.isArray(failingListCopy) || failingListCopy.length === 0)
			{
				console.log("There are no active elements to retry, aborting.")
				break;
			}

			for (const i of failingListCopy)
			{
				if (requestTasks.length >= concurrencyLimit)
				{
					await Promise.race(requestTasks);
				}

				console.log(`Retrying asset [${fileId}] ${i.source}`);

				if (cancelAcquisition)
				{
					break;
				}

				await fs.promises.mkdir(path.dirname(i.target), { recursive: true });
				const oneRequest = doRequest(session, i.target, i.source).then(() => 
				{
					requestTasks.splice(requestTasks.indexOf(oneRequest), 1);
				}).catch(() => 
				{
					requestTasks.splice(requestTasks.indexOf(oneRequest), 1);
				});
				requestTasks.push(oneRequest);
			}

			await Promise.all(requestTasks);

			if (imagesRequiringRetry && imagesRequiringRetry.length === 0)
			{
				console.log(`Retry ${i + 1} acquired all remaining files.`);
				break;
			}
		}
	}

	if (imagesRequiringRetry && imagesRequiringUpdate.length > 0)
	{
		console.log(`Retrying failed to acquire erroring files.`);
	}

	if (session)
	{
		closeSession(session);
		session = null;
	}

	console.log("Requests complete.");

	if (cancelAcquisition)
	{
		cancelAcquisition = false;
		downloadInProgress = false;

		// If we cancel, we'll have files we already downloaded so we should clear the cached list and recreate it
		imagesRequiringUpdate = null;
		numTransferred = 0;
		checkImageManifest();
	}
	else
	{
		downloadInProgress = false;
		win.webContents.send("acquireImages:downloadComplete");
	}
}

async function handleCancelImages()
{
	if (downloadInProgress)
	{
		cancelAcquisition = true;
	}
}

async function getFileHash(path, alg = "md5")
{
	return new Promise((resolve, __reject) => 
	{
		let hash = crypto.createHash(alg);
		let stream = fs.createReadStream(path);
		stream.on("error", __err => resolve(false));
		stream.on("data", chunk => hash.update(chunk));
		stream.on("end", () => resolve(hash.digest("base64")));
	});
}

async function checkImageManifest()
{
	/*
	If we already have an installation sentinel in place, assume we have all the files already.
	This avoids us needing to deal with newer versions of the files ending up on the server, or the hashes mis-matching.
	*/
	let installationStatus = await isInstalled();
	if (installationStatus)
	{
		hasImagesAvailableForUpdate(0, 0);
		return;
	}

	console.log("Querying image manifest for updated/missing files...");

	// Don't do anything if we're actively downloading things.
	if (downloadInProgress)
	{
		console.log("Bailing because download in progress.")
		return;
	}

	let manifestPath = path.join(app.getAppPath(), "image-manifest.json");

	if (!fs.existsSync(manifestPath))
	{
		console.log("Couldn't find image manifest file at:", manifestPath);
		return;
	}

	let requiredFileUpdates = [];
	let totalSize = 0;

	let images = fs.readFileSync(manifestPath, "utf8");
	images = JSON.parse(images);

	for (const i of images)
	{
		for (const s of packSizes)
		{
			let basePath = i.paths[s];
			if (process.platform === "linux" || process.platform === "darwin")
			{
				basePath = basePath.replaceAll("\\", "/");
			}
			
			let storedFilePath = path.normalize(path.join(app.getPath("userData"), "ImagePack", basePath));
			let remoteFilePath = `${baseURI}${i.paths[s].replaceAll("\\", "/")}`

			const fileHash = await getFileHash(storedFilePath, "md5");

			if (fileHash !== i.hashes[s])
			{
				requiredFileUpdates.push({
					source: remoteFilePath,
					target: storedFilePath
				});
				totalSize += i.sizes[s];
			}
		}
	}

	imagesRequiringUpdate = requiredFileUpdates;
	hasImagesAvailableForUpdate(requiredFileUpdates.length, totalSize);
}

async function handleDeleteImages()
{
	let targetPath = path.join(app.getAppPath(), "resources");

	// Remove the sentinel file. If we do nothing else this will make the game believe that the imagepack is no longer actually installed.
	let sentinelPath = path.join(targetPath, "installed.json");

	if (fs.existsSync(sentinelPath))
	{
		fs.rmSync(sentinelPath);
	}

	// Clear the cache folder entirely.
	let folderPath = path.join(app.getPath("userData"), "ImagePack");

	if (fs.existsSync(folderPath))
	{
		fs.rmdirSync(folderPath, { recursive: true });
	}

	// Clear out the merged folders
	const imgPath = path.join(targetPath, "img");
	let parents = fs.readdirSync(imgPath);
	
	parents.forEach(p => 
	{
		packSizes.forEach(s => 
		{
			let pathToDelete = path.join(imgPath, p, s);
			if (fs.existsSync(pathToDelete))
			{
				fs.rmdirSync(path.join(imgPath, p, s), { recursive: true });
			}
		});
	});
}

async function hasImagesAvailableForUpdate(numImages = 0, projectedSize = 0)
{
	console.log(`Sending message to render process for ${numImages} images`);
	win.webContents.send("acquireImages:notifyAvailable", numImages, projectedSize);
}

async function handleInstallImages()
{
	let fromPath = path.join(app.getPath("userData"), "ImagePack", "resources");
	let targetPath = path.join(app.getAppPath(), "resources");

	await fs.promises.cp(fromPath, targetPath, { recursive: true });
	let currentManifestHash = await getFileHash(path.join(app.getAppPath(), "image-manifest.json"));

	fs.writeFileSync(path.join(targetPath, "installed.json"), JSON.stringify({ installed: true, manifestHash: currentManifestHash }));
	win.webContents.send("acquireImages:notifyIsInstalled", true);
}

async function isInstalled()
{
	console.log("Testing if installed...");

	let markerPath = path.join(app.getAppPath(), "resources", "installed.json");

	if (!fs.existsSync(markerPath))
	{
		console.log(`Unable to find sentinel file.`);
		return false;
	}

	let marker = JSON.parse(fs.readFileSync(markerPath));

	let matchingInstalled = false;

	if (marker.manifestHash !== undefined)
	{
		let currentHash = await getFileHash(path.join(app.getAppPath(), "image-manifest.json"));

		console.log(`Installed hash: ${marker.manifestHash} Packaged manifest hash: ${currentHash}`);

		if (currentHash === marker.manifestHash)
		{
			matchingInstalled = true;
		}
	}

	console.log("Main is installed?", markerPath);

	return matchingInstalled;
}